%{
/* cfgparse.y -- Parse the configuration file.
   $Id: cfgparse.y,v 0.1 1998/07/04 16:57:35 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1998 by the Board of Trustees, Leland Stanford Jr. University

   Parse the configuration file.  Our end product is a linked list of log
   classes, which the rest of newsyslog will use as data to do the actual
   work (see cfgparse.h for the definitions of all of the relevant structs).
   The input stream comes from the lexer.

   Be warned:  We're pretty much relying on bison here because it makes the
   make process so much easier.  I don't think I'm doing anything
   bison-specific in this, but I'm also not testing against yacc regularly.
   So you may have to use bison. */

/***************************************************************************
* Includes and prototypes
***************************************************************************/

#include "expand.h"             /* expand() */
#include "logs.h"               /* Global struct definitions */
#include "utility.h"            /* xmalloc() */

#include <grp.h>                /* getgrnam() */
#include <pwd.h>                /* getpwnam() */
#include <stdarg.h>             /* va_list, va_start(), va_end() */
#include <stdio.h>              /* FILE, fopen(), fclose() */
#include <stdlib.h>             /* free(), exit() */
#include <string.h>             /* strcmp(), strcpy(), strlen() */

/* We want to have verbose parse error messages. */
#define YYERROR_VERBOSE 1

/* These two functions need to be externally visible. */
void cfgdie (char *format, ...);
void yyerror (char *error);
int yylex (); 

/* Stuff for internal use only. */
static struct archive *archive (char *template, int count);
static char *cfgexpand (char *string, char *name);
static char *lookup (char *path);
static void set (char *name, char *content);
static struct log *merge (struct log *logs, struct log *actions);


/***************************************************************************
* Global and external variables
***************************************************************************/
 
/* The FILE that our lexer will read from. */
extern FILE *yyin;

/* The current line of our configuration file (maintained by the lexer). */
extern int cfgline; 

/* Struct to store variable content mappings from set commands. */
struct variable {
  char *name;
  char *content;
  struct variable *next;
};

/* Stores a linked list of all variables that have been set. */
static struct variable *variables = NULL;
  
/* The final result of our parsing, a linked list of classes. */
static struct class *class;
%}

/***************************************************************************
* yacc declarations
***************************************************************************/

%start file

%union {
  char *string;
  int number;
  struct command *command;
  struct process *process;
  struct log *log;
  struct class *class;
}

%token ARCHIVE COPY EOL HUP LOG RESTART RUN SET STOP
%token EQ "="
%token LBRACE "{"
%token RBRACE "}"
%token <string> LABEL STRING VALUE VARIABLE
%token <number> NUMBER

%type <command> arguments
%type <process> stop restart process hup run
%type <number>  uid gid
%type <log>     logs log actions action archive copy
%type <class>   class block

%%

set             : SET VARIABLE "=" STRING EOL { set ($2, $4); }
                ;

hup             : HUP STRING
                  {
                    struct process *process;
                    
                    process = xmalloc (sizeof (struct process));
                    process->command = NULL;
                    process->pidfile = cfgexpand ($2, "");
                    $$ = process;
                  }
                ;

arguments       : arguments STRING
                  {
                    struct command *command = $1;
                    int size;

                    /* Check to see if we need to grow our array. */
                    command->argc++;
                    if (command->argc > command->size)
                      {
                        command->size *= 2;
                        size = command->size * sizeof (char *);
                        command->argv = xrealloc (command->argv, size);
                      }
                    command->argv[command->argc - 1] = cfgexpand ($2, "");
                    command->argv[command->argc] = NULL;
                  }
                | /* empty */
                  {
                    struct command *command;
                    
                    command = xmalloc (sizeof (struct command));
                    command->argc = 0;
                    command->size = 3;
                    command->argv = (char **) xmalloc (3 * sizeof (char *));
                    $$ = command;
                  }
                ;
                              
run             : RUN arguments
                {
                  struct process *process;
                  
                  process = xmalloc (sizeof (struct process));
                  process->command = $2;
                  process->pidfile = NULL;
                  $$ = process;
                }
                ;

process         : hup { $$ = $1; }
                | run { $$ = $1; }
                ;

stop            : STOP process EOL { $$ = $2; }
                | /* empty */ { $$ = NULL; }
                ;

restart         : RESTART process EOL { $$ = $2; }
                | /* empty */ { $$ = NULL; }
                ;

uid             : NUMBER { $$ = $1; }
                | STRING
                  {
                    struct passwd *pw;

                    pw = getpwnam ($1);
                    if (pw)
                      {
                        free ($1);
                        $$ = pw->pw_uid;
                      }
                    else
                      cfgdie ("Unknown user %s", $1);
                  }
                ;

gid             : NUMBER { $$ = $1; }
                | STRING
                  {
                    struct group *gr;

                    gr = getgrnam ($1);
                    if (gr)
                      {
                        free ($1);
                        $$ = gr->gr_gid;
                      }
                    else
                      cfgdie ("Unknown group %s", $1);
                  }
                ;

log             : STRING
                  {
                    struct log *log;

                    log = xmalloc (sizeof (struct log));
                    log->name = $1;
                    log->rotate = 0;
                    log->archive = NULL;
                    $$ = log;
                  }
                ;

logs            : LOG log STRING uid gid NUMBER EOL logs
                  {
                    struct log *log;
                    int mode, factor;

                    log = $2;
                    log->path = cfgexpand (lookup ($3), log->name);
                    log->tmppath = NULL;
                    log->owner = (uid_t) $4;
                    log->group = (gid_t) $5;
                    mode = $6;

                    /* Interpret the permissions as octal. */
                    factor = 1;
                    log->mode = 0;
                    while (mode > 0)
                      {
                        log->mode += (mode % 10) * factor;
                        factor *= 8;
                        mode /= 10;
                      }

                    /* Return a linked list of what we've seen so far. */
                    log->next = $8;
                    $$ = log;
                  }
                | /* empty */ { $$ = NULL; }
                ;

archive         : ARCHIVE log STRING NUMBER EOL
                  {
                    struct log *log;

                    log = $2;
                    log->archive = archive ($3, $4);
                    log->rotate = 1;
                    $$ = log;
                  }
                ;

copy            : COPY log STRING EOL
                  {
                    struct log *log;

                    log = $2;
                    log->archive = archive ($3, -1);
                    $$ = log;
                  }
                ;

action          : archive { $$ = $1; }
                | copy { $$ = $1; }
                ;

actions         : action actions
                  {
                    struct log *log;
                    
                    log = $1;
                    log->next = $2;
                    $$ = log;
                  }
                | /* empty */ { $$ = NULL; }
                ;

block           : LABEL "{" stop restart logs actions "}"
                  {
                    struct class *block;

                    block = xmalloc (sizeof (struct class));
                    block->name = $1;
                    block->stop = $3;
                    block->restart = $4;
                    block->logs = merge ($5, $6);
                    $$ = block;
                  }
                ;

class           : set class { $$ = $2; }
                | block class
                  {
                    struct class *block;

                    block = $1;
                    block->next = $2;
                    $$ = block;
                  }
                | /* empty */ { $$ = NULL; }
                ;

file            : class { class = $1; }
                ;

%%

/***************************************************************************
* Error handling
***************************************************************************/

/* yyerror() is just a simple wrapper around cfgdie(). */
void yyerror (char *error) { cfgdie (error); }


/* Prepend the current line number to the error message before dying.  This
   needs to not be a static function so that the lexer can pull it in and
   use it as well. */

void
cfgdie (char *format, ...)
{
  va_list message;

  fprintf (stderr, "config line %d: ", cfgline);
  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, "\n");
  exit (1);
}

   
/***************************************************************************
* Variable definition and lookup
***************************************************************************/

/* We maintain a linked list of variables that have been set, which is used
   later by lookup() when it does variable expansion.  Note that we don't
   check for duplicates; this is intentional, as we want to allow people to
   redefine variables.  We add all new definitions onto the *beginning* of
   the linked list, and then any subsequent lookups will always find the
   latest definition of a variable. */

static void
set (char *name, char *content)
{
  struct variable *variable;

  /* Allocate a new struct variable and add it to the linked list. */
  variable = xmalloc (sizeof (struct variable));
  variable->name = name;
  variable->content = content;
  variable->next = NULL;
  if (variables) variable->next = variables;
  variables = variable;
}


/* Given an argument that's supposed to be a path, run it through variable
   expansion if it doesn't start with a slash.  We're taking in a malloc'd
   string that we don't want if it turns out that it maps to a variable
   value, so free our argument if we're returning something out of the
   variable linked list. */

static char *
lookup (char *path)
{
  struct variable *variable;
  char *value;
  
  if (path[0] != '/')
    {
      for (variable = variables; variable; variable = variable->next)
        if (!strcmp (variable->name, path)) break;
      if (!variable) cfgdie ("Undefined variable %s", path);
      free (path);
      value = xmalloc (strlen (variable->content) + 1);
      strcpy (value, variable->content);
      return value;
    }
  else
    return path;
}


/***************************************************************************
* String expansion
***************************************************************************/

/* Check over a string for the invalid %s escape and then pass it along to
   expand() so that % expansion can be done.  The time passed to it will be
   0 (indicating to use the current time).  Returns NULL on error.  We free
   the string passed in and return the malloc'd string that expand()
   returns. */

static char *
cfgexpand (char *string, char *name)
{
  char *p;

  /* Check for invalid % escapes. */
  for (p = string; *p; p++)
    if (*p == '%' && !(*(++p) && strchr ("%dmnDMY", *p)))
      cfgdie ("Invalid %% escape");

  /* Okay, valid format.  Send it through the real expand(). */
  p = expand (string, name, 0);
  if (!p) cfgdie ("Unable to expand string");
  free (string);
  return p;
}


/***************************************************************************
* Structure generation
***************************************************************************/

/* Construct an archive struct, given a template or variable containing a
   template and the number of old logs to save as arguments.  We allocate a
   new struct archive and return it. */

static struct archive *
archive (char *template, int count)
{
  struct archive *archive;
  char *location;
  int length;

  /* Allocate a new archive struct and initialize it with the easy stuff. */
  archive = xmalloc (sizeof (struct archive));
  archive->count = count;
  archive->path = NULL;
  archive->next = NULL;

  /* Now massage the template according to what we're going to be doing. */
  location = lookup (template);
  length = strlen (location) + 4;
  if (count > 0) length += 2;
  archive->template = xmalloc (length);
  strcpy (archive->template, location);
  if (count > 0) strcat (archive->template, ".0");
  strcat (archive->template, ".gz");
  free (location);

  /* Return the completed archive struct. */
  return archive;
}


/* When we originally generated the list of logs and the list of actions, we
   did so as two separate linked lists of logs, as it was easier to convince
   yacc/bison to cope with that.  We now need to merge those two lists back
   together.  We take two pointers to linked lists of logs; the first will
   be correct for path, owner, group, and mode, and the second will have
   archives and rotate elements filled in.  We return a single pointer to a
   linked list of logs with no duplicates that has everything merged
   together.

   We have not previously checked to see if we have any duplicate log
   definitions, so we also check this while we're at it.  We call cfgdie()
   on any errors (rather than yyerror(), which can't take variable
   arguments). */

static struct log *
merge (struct log *logs, struct log *actions)
{
  struct log *current, *log, *old;
  struct archive *last;
  
  /* First, let's scan logs and make sure that there aren't any duplicates.
     We'll do this the long (simple) way. */
  for (current = logs; current->next; current = current->next)
    for (log = current->next; log; log = log->next)
      if (!strcmp (current->name, log->name))
        cfgdie ("Duplicate log definition for %s", current->name);

  /* Now, step through the actions one at a time, merging their contents
     into logs and freeing the memory used by the actions. */
  for (current = actions; current; )
    {
      /* Find the corresponding log. */
      for (log = logs; log; log = log->next)
        if (!strcmp (log->name, current->name)) break;

      /* Make sure we actually found a log. */
      if (!log) cfgdie ("Unknown log %s", current->name);

      /* Merge the actions over. */
      if (log->archive)
        {
          for (last = log->archive; last->next; last = last->next)
            ;
          last->next = current->archive;
        }
      else
        log->archive = current->archive;
      log->rotate |= current->rotate;

      /* Free the memory used by the action log struct. */
      old = current;
      current = current->next;
      free (old->name);
      free (old);
    }

  /* Return the linked list of combined log structs. */
  return logs;
}


/***************************************************************************
* Interface
***************************************************************************/

/* This is the function that the rest of the world gets to call.  It takes
   in the name of the configuration file, and returns a pointer to a linked
   list of class structs. */

struct class *
read_config (char *filename)
{
  FILE *config;

  /* Attempt to open our configuration file. */
  config = fopen (filename, "r");
  if (!config) die ("Cannot open configuration file %s", filename);

  /* Point our lexer at the right file, and then let magic happen. */
  yyin = config;
  if (yyparse () || !class) die ("Parsing of %s failed", filename);
  fclose (config);
  return class;
}
