/* file.c -- Functions to manipulate (copy, move, etc.) files.
   $Id: file.c,v 0.8 1998/07/04 17:43:28 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997, 1998 Board of Trustees, Leland Stanford Jr. University

   A set of functions to rename, copy, and create log files, including doing
   some specialized things we want to do because of the sorts of files that
   we're moving. */

/***************************************************************************
* Includes and prototypes
***************************************************************************/

#include "config.h"             /* COMPRESS */
#include "file.h"               /* Prototypes */
#include "process.h"            /* run() */
#include "utility.h"            /* xmalloc(), syswarn() */

#include <errno.h>              /* errno, EEXIST */
#include <fcntl.h>              /* open(), O_CREAT, O_EXCL, O_WRONLY */
#include <stdio.h>              /* sprintf(), rename(), fopen(), etc. */
#include <stdlib.h>             /* free() */
#include <string.h>             /* strlen(), strrchr() */
#include <sys/stat.h>           /* stat(), fstat(), fchmod(), mkdir() */
#include <sys/types.h>          /* off_t, uid_t, gid_t, mode_t */
#include <time.h>               /* time() */
#include <unistd.h>             /* fchown(), close(), access() */
#include <zlib.h>               /* gzdopen(), gzwrite(), gzclose() */

static void create_directory (char *path);

/***************************************************************************
* Renaming files
***************************************************************************/

/* Rename an existing log file to something else so that we can rebind the
   daemon to a new log file.  Logs are always renamed to <log_name>.<time>
   where <time> is the current time in seconds since the epoch.  This should
   reduce our chances of namespace collision to virtually nil, allowing the
   existence of the destination name to be a fatal error.  We return a
   pointer to the new name of the log on success and NULL on failure; memory
   for the new name is malloc()ed and the caller is responsible for freeing
   it. */

char *
rename_log (char *filename)
{
  char *newname;

  /* Allow 12 characters for time (a year 33679 problem!). */
  newname = (char *) xmalloc (strlen (filename) + 13);
  sprintf (newname, "%s.%lu", filename, (unsigned long) time (NULL));
  if (rename (filename, newname) != 0)
    {
      syswarn ("Rename of %s failed", filename);
      free (newname);
      return NULL;
    }
  return newname;
}


/* Do the magic of keeping a number of old logs, numbered consecutively
   (such as log.0.gz, log.1.gz, log.2.gz, etc.).  Delete the highest
   numbered log and then rename all logs numbered lower than that to make
   room for a new .0 log. */
int
shuffle_log (char *filename, int count)
{
  char *logname, *newlogname, *separator, *base;
  int i, remainder;
  int length = 0;

  /* Allocate enough space to store the largest possible rotation number.
     We don't want to require the math library just to get log10(), so we do
     it ourselves. */
  remainder = count;
  while (remainder > 0)
    {
      length++;
      remainder /= 10;
    }
  logname = (char *) xmalloc (strlen (filename) + length);
  base = (char *) xmalloc (strlen (filename));

  /* Build the base name, which is the filename with .0.gz stripped off. */
  strcpy (logname, filename);
  separator = strrchr (logname, '.');
  if (separator != NULL) *separator = '\0';
  if (separator != NULL) separator = strrchr (logname, '.');
  if (separator != NULL) *separator = '\0';
  if (separator == NULL)
    {
      warn ("Invalid archive name %s", filename);
      free (base);
      free (logname);
      return 0;
    }
  strcpy (base, logname);

  /* Remove the highest numbered log, if it exists. */
  sprintf (separator, ".%d.gz", count - 1);
  if (access (logname, F_OK) == 0)
    if (unlink (logname) != 0)
      {
        syswarn ("Could not delete %s", logname);
        free (base);
        free (logname);
        return 1;
      }

  /* Step through each log numbered smaller than that and rename it to the
     next highest number. */
  newlogname = (char *) xmalloc (strlen (filename) + length);
  for (i = count - 2; i >= 0; i--)
    {
      sprintf (logname, "%s.%d.gz", base, i);
      sprintf (newlogname, "%s.%d.gz", base, i + 1);
      if (access (logname, F_OK) == 0)
        if (rename (logname, newlogname) != 0)
          {
            syswarn ("Unable to rename %s", logname);
            free (base);
            free (logname);
            free (newlogname);
            return 0;
          }
    }

  /* All done.  Clean up the memory we allocated. */
  free (base);
  free (logname);
  free (newlogname);
  return 1;
}


/***************************************************************************
* Copying and moving files
***************************************************************************/

/* Copy a file from one name file to another.  We do this ourselves rather
   than forking a cp process so that we can be sure of the error checking.
   Returns 1 on success, 0 on failure. */

int
copy_log (char *sourcename, char *destname)
{
  FILE *source, *dest;
  struct stat statbuf;
  int nread, nwritten, have_size;
  off_t size = 0;
  mode_t mode = 0;

  /* Buffer for transferring the file. */
  static char buffer[2048];

  /* Open the two files. */
  source = fopen (sourcename, "rb");
  if (source == NULL)
    {
      syswarn ("Cannot open source file %s", sourcename);
      return 0;
    }
  dest = fopen (destname, "wb");
  if (dest == NULL)
    {
      warn ("Cannot create %s", destname);
      return 0;
    }

  /* Get the original size of the file in bytes. */
  if (fstat (fileno (source), &statbuf) != 0)
    {
      syswarn ("Unable to stat source file %s", sourcename);
      have_size = 0;
    }
  else
    {
      size = statbuf.st_size;
      mode = statbuf.st_mode;
      have_size = 1;
    }

  /* If we can, chmod the archived log to have the same permissions as the
     original. */
  if (have_size)
    if (fchmod (fileno (dest), mode) != 0)
      syswarn ("Unable to chmod destination file %s", destname);

  /* Copy the file over. */
  while (!feof (source) && !ferror (source))
    {
      nread = fread (buffer, 1, 2047, source);
      nwritten = fwrite (buffer, 1, nread, dest);
      if (nread != nwritten)
        {
          warn ("Error while writing %s", destname);
          fclose (source);
          fclose (dest);
          return 0;
        }
    }
  if (ferror (source)) warn ("Error while reading %s", sourcename);

  /* Close the files and then verify that the copy was successful. */
  fclose (source);
  if (fclose (dest) != 0)
    {
      warn ("Error while writing %s", destname);
      return 0;
    }
  if (!have_size || stat (destname, &statbuf) != 0)
    warn ("Unable to verify successful copy of %s", sourcename);
  else if (size != statbuf.st_size)
    {
      warn ("Error copying %s", sourcename);
      warn ("  size of copy (%lu) != size of original (%lu)",
            statbuf.st_size, size);
      return 0;
    }

  return 1;
}


/* Acts just like compress_log(), but allows overwriting of an existing
   destination file. */

int
overwrite_log (char *sourcename, char *destname)
{
  /* Create the directory, if necessary. */
  create_directory (destname);

  /* Unlink the file if it already exists and then try to copy the log. */
  if (access (destname, F_OK) == 0)
    if (unlink (destname) != 0)
      {
        syswarn ("Could not remove %s", destname);
        return 0;
      }
  return compress_log (sourcename, destname);
}


/* Acts just like compress_log(), but also attempts to create the directory
   for destname if it doesn't already exist.  We should allow the
   permissions on the directory to be passed in somehow.... */

int
archive_log (char *sourcename, char *destname)
{
  /* Create the directory, if necessary. */
  create_directory (destname);

  /* Try to copy the log.  This will fail if the directory doesn't exist. */
  if (!compress_log (sourcename, destname)) return 0;

  return 1;
}


/***************************************************************************
* Creating files and directories
***************************************************************************/

/* Create a new log file with the given name, owner, group, and permissions.
   Returns 1 on success and 0 on failure. */

int
create_log (char *filename, uid_t owner, gid_t group, mode_t mode)
{
  int fd;

  /* Create the file. */
  fd = open (filename, O_CREAT | O_WRONLY, mode);
  if (fd < 0)
    {
      syswarn ("Creation of %s failed", filename);
      return 0;
    }

  /* chown the new log to the right owner and group. */
  if (fchown (fd, owner, group) != 0)
    syswarn ("Unable to chown %s", filename);

  /* chmod the new log to the right file permissions. */
  if (fchmod (fd, mode) != 0)
    syswarn ("Unable to chmod %s", filename);
  
  close (fd);
  return 1;
}


/* Creates the last directory in a given path if it doesn't already exist,
   used for both archive_log() and overwrite_log() so that paths can be used
   that create a new directory each day. */

static void
create_directory (char *path)
{
  char *separator;

  /* Find the last slash, if it exists. */
  separator = strrchr (path, '/');
  if (separator != NULL)
    {
      /* Change it to a null temporarily. */
      *separator = '\0';

      /* Check for the existence of the directory. */
      if (access (path, W_OK) != 0)
        if (mkdir (path, 0755) != 0 && errno != EEXIST)
          syswarn ("Could not create directory %s", path);

      /* Change the last slash back. */
      *separator = '/';
    }
}


/***************************************************************************
* Compressing logs
***************************************************************************/

/* Copy a file from one location to another, compressing it in the process.
   The preferred method for doing so is to link against zlib and use the
   zlib functions to write a gzipped file.  If we can't do that, we'll have
   to use a pipe to a gzip process (not currently implemented). */

int
compress_log (char *sourcename, char *destname)
{
  FILE *source;
  gzFile *dest;
  struct stat statbuf;
  int fd, nread, nwritten, have_mode;

  /* Buffer for transferring the file. */
  static char buffer[2048];

  /* Open the source file as a standard file and get its mode. */
  source = fopen (sourcename, "rb");
  if (source == NULL)
    {
      syswarn ("Cannot open source file %s", sourcename);
      return 0;
    }
  if (fstat (fileno (source), &statbuf) != 0)
    {
      syswarn ("Unable to stat source file %s", sourcename);
      have_mode = 0;
    }
  else
    have_mode = 1;

  /* Open the destination file and try to chmod it to the right perms. */
  fd = open (destname, O_CREAT | O_EXCL | O_WRONLY, statbuf.st_mode);
  if (fd < 0)
    {
      syswarn ("Cannot create %s", destname);
      fclose (source);
      return 0;
    }
  if (have_mode)
    if (fchmod (fd, statbuf.st_mode) != 0)
      syswarn ("Unable to chmod destination file %s", destname);

  /* Now, open this file descriptor as a maximally compressed file. */
  dest = gzdopen (fd, "wb9");
  if (dest == NULL)
    {
      warn ("Cannot begin compressing %s", destname);
      fclose (source);
      close (fd);
      return 0;
    }

  /* Copy the file over.  Maybe some day add better error reporting. */
  while (!feof (source) && !ferror (source))
    {
      nread = fread (buffer, 1, 2047, source);
      nwritten = gzwrite (dest, buffer, nread);
      if (nread != nwritten)
        {
          warn ("Error while writing %s", destname);
          fclose (source);
          gzclose (dest);
          return 0;
        }
    }
  if (ferror (source)) warn ("Error while reading %s", sourcename);

  /* Close the files and then verify that the copy was successful. */
  fclose (source);
  if (gzclose (dest) != 0)
    {
      warn ("Error while writing %s", destname);
      return 0;
    }

  return 1;
}
