/* main.c -- Log file rotation and periodic maintenance.
   $Id: main.c,v 0.5 1998/07/04 17:11:29 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997, 1998 Board of Trustees, Leland Stanford Jr. University

   This program is designed to rotate, truncate, or otherwise deal with
   system logs on a regular basis, including correctly stopping and
   restarting daemons so that the logs can be moved and compressed safely.
   It takes a configuration file specifying which logs to rotate and what to
   do with them. */

/***************************************************************************
* Includes
***************************************************************************/

#include "logs.h"               /* read_config(), struct class, etc. */
#include "config.h"             /* CONFIG_FILE */
#include "expand.h"             /* expand() */
#include "file.h"               /* rename_log(), copy_log(), etc. */
#include "process.h"            /* run(), hup() */
#include "utility.h"            /* warn() */

#include <stdlib.h>             /* free(), exit() */
#include <unistd.h>             /* unlink() */


/***************************************************************************
* Running commands
***************************************************************************/

/* Deal with a process (from a struct process), which involves either
   sending SIGHUP to a process from a PID file or running a command with a
   variable number of arguments. */

static void
process (struct process *process)
{
  if (process->pidfile)
    hup (process->pidfile);
  else if (process->command)
    run (process->command->argv);
}


/***************************************************************************
* Main routine
***************************************************************************/

/* Read in our configuration file, getting the list of log classes, and then
   step through each one, performing the desired tasks. */

int
main (int argc, char **argv)
{
  int okay;
  struct class *class, *current;
  struct log *log;
  struct archive *archive;

  /* If we have a command-line argument, assume that's the config file. */
  if (argc > 1)
    class = read_config (argv[1]);
  else
    class = read_config (CONFIG_FILE);

  /* Step through each class and perform the desired operations. */
  for (current = class; current; current = current->next)
    {
      /* Stop the daemon if necessary. */
      if (current->stop) process (current->stop);

      /* Rotate each log and create the new log file. */
      for (log = current->logs; log; log = log->next)
        {
          if (!log->rotate) continue;
          log->tmppath = rename_log (log->path);
          if (log->tmppath)
            create_log (log->path, log->owner, log->group, log->mode);
        }

      /* Restart the daemon, if necessary. */
      if (current->restart) process (current->restart);

      /* Copy each log to each of its archive locations and compress it. */
      for (log = current->logs; log; log = log->next)
        {
          if ((!log->tmppath && log->rotate) || !log->archive)
            warn ("Skipping archiving of %s", log->name);
          else
            {
              okay = 1;
              for (archive = log->archive; archive; archive = archive->next)
                {
                  archive->path = expand (archive->template, log->name, 0);
                  if (archive->count > 0)
                    okay &= shuffle_log (archive->path, archive->count);
                  if (archive->count >= 0)
                    okay &= archive_log (log->tmppath, archive->path);
                  else
                    okay &= overwrite_log (log->path, archive->path);
                }
              if (log->rotate && okay && unlink (log->tmppath) != 0)
                syswarn ("Cannot delete temporary file %s");
            }
        }
    }

  /* Success.  Exit. */
  exit (0);
}
