/* process.c -- Functions for running and signalling processes.
   $Id: process.c,v 0.2 1998/02/02 20:19:49 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997 by the Board of Trustees, Leland Stanford Jr. University

   Some general utility functions for running and signalling processes.  We
   don't want to use system() to run processes due to problems with the
   shell, metacharacters, memory usage, and the like, so we have our own
   function which just does a fork and exec.  We also need to be able to
   signal programs with a HUP to rebind to new logs (syslog has this
   behavior, for example). */

/***************************************************************************
* Includes and prototypes
***************************************************************************/

#include "process.h"            /* Prototypes */
#include "utility.h"            /* warn(), syswarn() */

#include <signal.h>             /* kill(), SIGHUP */
#include <stdio.h>              /* fopen(), fscanf(), fclose(), FILE */
#include <stdlib.h>             /* free() */
#include <sys/types.h>          /* pid_t */
#include <sys/wait.h>           /* wait(), WIFEXITED() */
#include <unistd.h>             /* fork(), execv(), sleep() */


/***************************************************************************
* Program execution
***************************************************************************/

/* Execute a program with fork() and exec() and pass it an arbitrary number
   of arguments.  This is a very limited reimplimentation of something like
   execv() combined with system() which is designed to suit just our
   purposes.  Note that the argv array *has* to be NULL-terminated by the
   caller or evil things happen. */

int
run (char **argv)
{
  pid_t child;
  int status;

  /* Assume success, set to 0 on failure. */
  int success = 1;

  /* Fork off the process and execute it. */
  child = fork ();
  if (child == (pid_t) -1)
    {
      syswarn ("fork failed");
      success = 0;
    }
  else if (child == 0)
    {
      /* In child process, so exec the program. */
      execv (argv[0], argv);

      /* Should never be reached. */
      sysdie ("exec failed");
    }
  else
    {
      /* In parent, wait for child to terminate. */
      child = wait (&status);
      if (child < 0)
        {
          syswarn ("wait failed");
          success = 0;
        }
      if (!WIFEXITED (status))
        {
          warn ("%s terminated abnormally", argv[0]);
          success = 0;
        }
    }

  /* Return our success or lack thereof. */
  return success;
}


/***************************************************************************
* Signals
***************************************************************************/

/* Send a SIGHUP to a process, reading the PID of the process to SIGHUP from
   a file.  This should be general enough to deal with syslogd and most
   everything else that works similar to it.  Returns 1 on success, 0 on
   failure. */

int
hup (char *pid_filename)
{
  FILE *pid_file;
  int pid = 0;

  /* Read the PID from the given filename. */
  pid_file = fopen (pid_filename, "r");
  if (pid_file == NULL)
    {
      warn ("Cannot open PID file %s", pid_filename);
      return 0;
    }
  fscanf (pid_file, "%d", &pid);
  fclose (pid_file);

  /* Don't accept invalid PIDs and *never* send SIGHUP to init. */
  if (pid < 2)
    {
      warn ("PID file %s corrupt", pid_filename);
      return 0;
    }

  /* Send the signal. */
  if (kill ((pid_t) pid, SIGHUP) != 0)
    {
      syswarn ("HUP %d from PID file %s failed", pid, pid_filename);
      return 0;
    }

  /* Sleep a couple of seconds to give the daemon time to reset. */
  sleep (2);

  /* Make sure the process is still running. */
  if (kill (pid, 0) != 0)
    {
      syswarn ("Unable to verify process %d is running", pid);
      warn ("Daemon HUPed from file %s may have died", pid, pid_filename);
      return 0;
    }
  else
    return 1;
}
